//-----------------------------------------------------------------------------
// Copyright (c) 2008-2011 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

using Dhpoware;

namespace XNAThirdPersonCamera
{
    /// <summary>
    /// This Windows XNA application implements a third person style camera. A
    /// ThirdPersonCamera class is provided that implements the third person
    /// camera.
    /// 
    /// <para>The camera and the 3D model are attached to each other by a
    /// spring. As the user moves the 3D model around the scene the camera
    /// is dragged along with it. When the user stops moving the 3D model
    /// the camera slowly moves back to its resting position behind the 3D
    /// model.</para>
    /// 
    /// <para>The purpose of the spring system in the third person camera is
    /// to smooth out the camera movement. If you disable the camera spring
    /// system in the demo the camera will instantly snap to its new position
    /// in response to the user's input.</para>
    /// 
    /// <para>The behavior of the spring system is governed by the spring's
    /// stiffness. The less stiff the spring the more laggy the camera becomes.
    /// Increasing the spring's stiffness will make the camera less laggy. The
    /// correct spring stiffness will depend on your application. However very
    /// small stiffness values will cause the camera to lag too far behind the
    /// 3D model, and very large stiffness values will negate the spring
    /// damping effect on the camera.</para>
    /// </summary>
    public class Demo : Microsoft.Xna.Framework.Game
    {
        private static void Main()
        {
            using (Demo demo = new Demo())
            {
                demo.Run();
            }
        }

        /// <summary>
        /// A light. This light structure is the same as the one defined in the
        /// blinn_phong.fx file.
        /// </summary>
        private struct Light
        {
            public Vector3 Direction;
            public Vector3 Position;
            public Vector4 Ambient;
            public Vector4 Diffuse;
            public Vector4 Specular;
            public float SpotInnerConeRadians;
            public float SpotOuterConeRadians;
            public float Radius;
        }

        /// <summary>
        /// A material. This material structure is the same as the one defined
        /// in the blinn_phong.fx file.
        /// </summary>
        private struct Material
        {
            public Vector4 Ambient;
            public Vector4 Diffuse;
            public Vector4 Emissive;
            public Vector4 Specular;
            public float Shininess;
        }

        private const float BALL_FORWARD_SPEED = 120.0f;
        private const float BALL_HEADING_SPEED = 120.0f;
        private const float BALL_ROLLING_SPEED = 280.0f;

        private const float FLOOR_WIDTH = 1024.0f;
        private const float FLOOR_HEIGHT = 1024.0f;
        private const float FLOOR_TILE_U = 4.0f;
        private const float FLOOR_TILE_V = 4.0f;
        
        private const float CAMERA_FOVX = 80.0f;
        private const float CAMERA_ZFAR = FLOOR_WIDTH * 2.0f;
        private const float CAMERA_ZNEAR = 1.0f;
        private const float CAMERA_MAX_SPRING_CONSTANT = 100.0f;
        private const float CAMERA_MIN_SPRING_CONSTANT = 1.0f;
        
        private const float LIGHT_RADIUS = 1024.0f;
        private const float LIGHT_SPOT_INNER_CONE = 30.0f;
        private const float LIGHT_SPOT_OUTER_CONE = 100.0f;

        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private SpriteFont spriteFont;
        private Effect effect;
        private Model ballModel;
        private ThirdPersonCamera camera;
        private Entity ballEntity;
        private VertexBuffer vertexBuffer;
        private Texture2D floorColorMap;
        private Vector2 fontPos;
        private Vector4 globalAmbient;
        private Matrix[] modelTransforms;
        private Light light;
        private Material material;
        private KeyboardState currentKeyboardState;
        private KeyboardState prevKeyboardState;
        private int windowWidth;
        private int windowHeight;
        private int frames;
        private int framesPerSecond;
        private float ballRadius;
        private TimeSpan elapsedTime = TimeSpan.Zero;
        private bool displayHelp;

        public Demo()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Window.Title = "XNA 4.0 Third Person Camera Demo";
            IsMouseVisible = true;
            IsFixedTimeStep = false;
        }

        private void CreateFloor()
        {
            float w = FLOOR_WIDTH  * 0.5f;
            float h = FLOOR_HEIGHT * 0.5f;
            
            Vector3[] positions =
            {
                new Vector3(-w, 0.0f, -h),
                new Vector3( w, 0.0f, -h),
                new Vector3(-w, 0.0f,  h),
                new Vector3( w, 0.0f,  h)
            };

            Vector2[] texCoords =
            {
                new Vector2(0.0f,         0.0f),
                new Vector2(FLOOR_TILE_U, 0.0f),
                new Vector2(0.0f,         FLOOR_TILE_V),
                new Vector2(FLOOR_TILE_U, FLOOR_TILE_V)
            };
            
            VertexPositionNormalTexture[] vertices =
            {
                new VertexPositionNormalTexture(
                    positions[0], Vector3.Up, texCoords[0]),
                new VertexPositionNormalTexture(
                    positions[1], Vector3.Up, texCoords[1]),
                new VertexPositionNormalTexture(
                    positions[2], Vector3.Up, texCoords[2]),
                new VertexPositionNormalTexture(
                    positions[3], Vector3.Up, texCoords[3]),
            };
            
            vertexBuffer = new VertexBuffer(GraphicsDevice,
                typeof(VertexPositionNormalTexture), vertices.Length,
                BufferUsage.WriteOnly);

            vertexBuffer.SetData(vertices);
        }

        protected override void Initialize()
        {
            base.Initialize();
        
            // Setup the window to be a quarter the size of the desktop.
            windowWidth = GraphicsDevice.DisplayMode.Width / 2;
            windowHeight = GraphicsDevice.DisplayMode.Height / 2;

            // Setup frame buffer.
            graphics.SynchronizeWithVerticalRetrace = false;
            graphics.PreferredBackBufferWidth = windowWidth;
            graphics.PreferredBackBufferHeight = windowHeight;
            graphics.PreferMultiSampling = true;
            graphics.ApplyChanges();

            // Position the text.
            fontPos = new Vector2(1.0f, 1.0f);

            // Setup the initial input states.
            currentKeyboardState = Keyboard.GetState();

            // Setup a spot light.
            globalAmbient = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);
            light.Direction = Vector3.Down;
            light.Position = new Vector3(0.0f, LIGHT_RADIUS * 0.5f, 0.0f);
            light.Ambient = Color.White.ToVector4();
            light.Diffuse = Color.White.ToVector4();
            light.Specular = Color.White.ToVector4();
            light.SpotInnerConeRadians = MathHelper.ToRadians(LIGHT_SPOT_INNER_CONE);
            light.SpotOuterConeRadians = MathHelper.ToRadians(LIGHT_SPOT_OUTER_CONE);
            light.Radius = LIGHT_RADIUS;
            
            // Setup a Lambert material.
            material.Ambient = new Vector4(0.2f, 0.2f, 0.2f, 1.0f);
            material.Diffuse = new Vector4(0.8f, 0.8f, 0.8f, 1.0f);
            material.Emissive = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);
            material.Specular = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);
            material.Shininess = 0.0f;
        
            // Create floor geometry.
            CreateFloor();
        
            // Determine the radius of the ball model.           
            BoundingSphere bounds = new BoundingSphere();
            foreach (ModelMesh mesh in ballModel.Meshes)
                bounds = BoundingSphere.CreateMerged(bounds, mesh.BoundingSphere);
            ballRadius = bounds.Radius;
        
            // Setup the camera.
            camera = new ThirdPersonCamera();
            camera.Perspective(CAMERA_FOVX, (float)windowWidth / (float)windowHeight,
                CAMERA_ZNEAR, CAMERA_ZFAR);
            camera.LookAt(new Vector3(0.0f, ballRadius * 3.0f, ballRadius * 7.0f),
                Vector3.Zero, Vector3.Up);
                
            // Setup the ball entity.
            ballEntity = new Entity();
            ballEntity.ConstrainToWorldYAxis = true;
            ballEntity.Position = new Vector3(0.0f, 1.0f + ballRadius, 0.0f);
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            spriteFont = Content.Load<SpriteFont>(@"Fonts\DemoFont");
            floorColorMap = Content.Load<Texture2D>(@"Textures\floor_decal_map");
            effect = Content.Load<Effect>(@"Effects\blinn_phong");
            ballModel = Content.Load<Model>(@"Models\ball");
        }

        protected override void UnloadContent()
        {
        }

        private bool KeyJustPressed(Keys key)
        {
            return currentKeyboardState.IsKeyDown(key) && prevKeyboardState.IsKeyUp(key);
        }

        private void ProcessKeyboard()
        {
            prevKeyboardState = currentKeyboardState;
            currentKeyboardState = Keyboard.GetState();

            if (KeyJustPressed(Keys.Escape))
                this.Exit();

            if (KeyJustPressed(Keys.H))
                displayHelp = !displayHelp;

            if (KeyJustPressed(Keys.Space))
                camera.EnableSpringSystem = !camera.EnableSpringSystem;

            if (currentKeyboardState.IsKeyDown(Keys.LeftAlt) ||
                currentKeyboardState.IsKeyDown(Keys.RightAlt))
            {
                if (KeyJustPressed(Keys.Enter))
                    ToggleFullScreen();
            }
            
            if (KeyJustPressed(Keys.Add))
            {
                float springConstant = camera.SpringConstant + 0.1f;

                springConstant = Math.Min(CAMERA_MAX_SPRING_CONSTANT, springConstant);
                camera.SpringConstant = springConstant;
            }
            
            if (KeyJustPressed(Keys.Subtract))
            {
                float springConstant = camera.SpringConstant - 0.1f;

                springConstant = Math.Max(CAMERA_MIN_SPRING_CONSTANT, springConstant);
                camera.SpringConstant = springConstant;
            }
        }

        private void ToggleFullScreen()
        {
            int newWidth = 0;
            int newHeight = 0;

            graphics.IsFullScreen = !graphics.IsFullScreen;

            if (graphics.IsFullScreen)
            {
                newWidth = GraphicsDevice.DisplayMode.Width;
                newHeight = GraphicsDevice.DisplayMode.Height;
            }
            else
            {
                newWidth = windowWidth;
                newHeight = windowHeight;
            }

            graphics.PreferredBackBufferWidth = newWidth;
            graphics.PreferredBackBufferHeight = newHeight;
            graphics.ApplyChanges();

            camera.Perspective(CAMERA_FOVX, (float)newWidth / (float)newHeight,
                CAMERA_ZNEAR, CAMERA_ZFAR);
        }

        protected override void Update(GameTime gameTime)
        {
            if (!IsActive)
                return;

            ProcessKeyboard();
            UpdateBall(gameTime);
            UpdateFrameRate(gameTime);

            base.Update(gameTime);
        }

        private void UpdateBall(GameTime gameTime)
        {
            float pitch = 0.0f;
            float heading = 0.0f;
            float forwardSpeed = 0.0f;
            
            if (currentKeyboardState.IsKeyDown(Keys.W) || 
                currentKeyboardState.IsKeyDown(Keys.Up))
            {
                forwardSpeed = BALL_FORWARD_SPEED;
                pitch = -BALL_ROLLING_SPEED;
            }
            
            if (currentKeyboardState.IsKeyDown(Keys.S) ||
                currentKeyboardState.IsKeyDown(Keys.Down))
            {
                forwardSpeed = -BALL_FORWARD_SPEED;
                pitch = BALL_ROLLING_SPEED;
            }
            
            if (currentKeyboardState.IsKeyDown(Keys.D) ||
                currentKeyboardState.IsKeyDown(Keys.Right))
            {
                heading = -BALL_HEADING_SPEED;
            }
            
            if (currentKeyboardState.IsKeyDown(Keys.A) ||
                currentKeyboardState.IsKeyDown(Keys.Left))
            {
                heading = BALL_HEADING_SPEED;
            }

            // Prevent the ball from rolling off the edge of the floor.
            
            float floorBoundaryZ = FLOOR_HEIGHT * 0.5f - ballRadius;
            float floorBoundaryX = FLOOR_WIDTH * 0.5f - ballRadius;
            float elapsedTimeSec = (float)gameTime.ElapsedGameTime.TotalSeconds;
            float velocity = forwardSpeed * elapsedTimeSec;
            Vector3 newBallPos = ballEntity.Position + ballEntity.Forward * velocity;
            
            if (newBallPos.Z > floorBoundaryZ)
                forwardSpeed = 0.0f;
            else if (newBallPos.Z < -floorBoundaryZ)
                forwardSpeed = 0.0f;
            else if (newBallPos.X > floorBoundaryX)
                forwardSpeed = 0.0f;
            else if (newBallPos.X < -floorBoundaryX)
                forwardSpeed = 0.0f;
            
            // Update the ball's state.
            
            ballEntity.Velocity = new Vector3(0.0f, 0.0f, forwardSpeed);
            ballEntity.Orient(heading, 0.0f, 0.0f);
            ballEntity.Rotate(0.0f, pitch, 0.0f);
            ballEntity.Update(gameTime);

            // Then move the camera based on where the ball has moved to.
            // When the ball is moving backwards rotations are inverted to
            // match the direction of travel. Consequently the camera's
            // rotation needs to be inverted as well.
            
            camera.Rotate((forwardSpeed >= 0.0f) ? heading : -heading, 0.0f);
            camera.LookAt(ballEntity.Position);
            camera.Update(gameTime);
        }

        private void UpdateFrameRate(GameTime gameTime)
        {
            elapsedTime += gameTime.ElapsedGameTime;

            if (elapsedTime > TimeSpan.FromSeconds(1))
            {
                elapsedTime -= TimeSpan.FromSeconds(1);
                framesPerSecond = frames;
                frames = 0;
            }
        }

        private void IncrementFrameCounter()
        {
            ++frames;
        }

        private void DrawFloor()
        {
            GraphicsDevice.SetVertexBuffer(vertexBuffer);
            
            // Set shader matrix parameters.
            effect.Parameters["worldMatrix"].SetValue(Matrix.Identity);
            effect.Parameters["worldInverseTransposeMatrix"].SetValue(Matrix.Identity);
            effect.Parameters["worldViewProjectionMatrix"].SetValue(camera.ViewProjectionMatrix);

            // Set the shader camera position parameter.
            effect.Parameters["cameraPos"].SetValue(camera.Position);

            // Set the shader global ambiance parameters.
            effect.Parameters["globalAmbient"].SetValue(globalAmbient);

            // Set the shader lighting parameters.
            effect.Parameters["light"].StructureMembers["dir"].SetValue(light.Direction);
            effect.Parameters["light"].StructureMembers["pos"].SetValue(light.Position);
            effect.Parameters["light"].StructureMembers["ambient"].SetValue(light.Ambient);
            effect.Parameters["light"].StructureMembers["diffuse"].SetValue(light.Diffuse);
            effect.Parameters["light"].StructureMembers["specular"].SetValue(light.Specular);
            effect.Parameters["light"].StructureMembers["spotInnerCone"].SetValue(light.SpotInnerConeRadians);
            effect.Parameters["light"].StructureMembers["spotOuterCone"].SetValue(light.SpotOuterConeRadians);
            effect.Parameters["light"].StructureMembers["radius"].SetValue(light.Radius);

            // Set the shader material parameters.
            effect.Parameters["material"].StructureMembers["ambient"].SetValue(material.Ambient);
            effect.Parameters["material"].StructureMembers["diffuse"].SetValue(material.Diffuse);
            effect.Parameters["material"].StructureMembers["emissive"].SetValue(material.Emissive);
            effect.Parameters["material"].StructureMembers["specular"].SetValue(material.Specular);
            effect.Parameters["material"].StructureMembers["shininess"].SetValue(material.Shininess);
            
            // Bind the texture map to the shader.
            effect.Parameters["colorMapTexture"].SetValue(floorColorMap);
                       
            effect.CurrentTechnique = effect.Techniques["PerPixelSpotLighting"];
            
            foreach (EffectPass pass in effect.CurrentTechnique.Passes)
            {
                pass.Apply();
                GraphicsDevice.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);
            }
        }

        private void DrawBall()
        {
            if (modelTransforms == null)
                modelTransforms = new Matrix[ballModel.Bones.Count];
        
            ballModel.CopyAbsoluteBoneTransformsTo(modelTransforms);
                
            foreach (ModelMesh m in ballModel.Meshes)
            {
                foreach (BasicEffect e in m.Effects)
                {
                    e.PreferPerPixelLighting = true;
                    e.TextureEnabled = true;
                    e.EnableDefaultLighting();
                    e.World = modelTransforms[m.ParentBone.Index] * ballEntity.WorldMatrix;
                    e.View = camera.ViewMatrix;
                    e.Projection = camera.ProjectionMatrix; 
                }
                
                m.Draw();
            }
        }

        private void DrawText()
        {
            StringBuilder buffer = new StringBuilder();

            if (displayHelp)
            {
                buffer.AppendLine("Press W or UP to roll the ball forwards");
                buffer.AppendLine("Press S or DOWN to roll the ball backwards");
                buffer.AppendLine("Press D or RIGHT to turn the ball to the right");
                buffer.AppendLine("Press A or LEFT to turn the ball to the left");
                buffer.AppendLine();
                buffer.AppendLine("Press SPACE to enable and disable the camera's spring system");
                buffer.AppendLine("Press + and - to change the camera's spring constant");
                buffer.AppendLine("Press ALT and ENTER to toggle full screen");
                buffer.AppendLine("Press ESCAPE to exit");
                buffer.AppendLine();
                buffer.AppendLine("Press H to hide help");
            }
            else
            {
                buffer.AppendFormat("FPS: {0}\n", framesPerSecond);
                
                bool springOn = camera.EnableSpringSystem;
                float springConstant = camera.SpringConstant;
                float dampingConstant = camera.DampingConstant;

                buffer.AppendLine();
                buffer.AppendLine("Camera");
                buffer.AppendFormat("  Spring {0}\n", (springOn ? "enabled" : "disabled"));
                buffer.AppendFormat("  Spring constant: {0}\n", springConstant.ToString("f2"));
                buffer.AppendFormat("  Damping constant: {0}\n", dampingConstant.ToString("f2"));
                buffer.AppendLine();
                buffer.AppendLine("Press H to display help");
            }

            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.AlphaBlend);
            spriteBatch.DrawString(spriteFont, buffer.ToString(), fontPos, Color.Yellow);
            spriteBatch.End();
        }

        protected override void Draw(GameTime gameTime)
        {
            if (!IsActive)
                return;
        
            GraphicsDevice.Clear(Color.Black);

            GraphicsDevice.BlendState = BlendState.Opaque;
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.SamplerStates[0] = SamplerState.LinearWrap;

            DrawBall();
            DrawFloor();
            DrawText();

            base.Draw(gameTime);
            IncrementFrameCounter();
        }
    }
}